

#include "jackvst.h"

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkevents.h>
#include <X11/Xlib.h>

#define GDK_PIXMAP_SCREEN(pix)	      (GDK_DRAWABLE_IMPL_X11 (((GdkPixmapObject *)pix)->impl)->screen)
#define GDK_PIXMAP_DISPLAY(pix)       (GDK_SCREEN_X11 (GDK_PIXMAP_SCREEN (pix))->display)
#define GDK_PIXMAP_XROOTWIN(pix)      (GDK_SCREEN_X11 (GDK_PIXMAP_SCREEN (pix))->xroot_window)
#define GDK_DRAWABLE_DISPLAY(win)     (GDK_IS_WINDOW (win) ? GDK_WINDOW_DISPLAY (win) : GDK_PIXMAP_DISPLAY (win))
#define GDK_DRAWABLE_SCREEN(win)      (GDK_IS_WINDOW (win) ? GDK_WINDOW_SCREEN (win) : GDK_PIXMAP_SCREEN (win))
#define GDK_DRAWABLE_XROOTWIN(win)    (GDK_IS_WINDOW (win) ? GDK_WINDOW_XROOTWIN (win) : GDK_PIXMAP_XROOTWIN (win))
#define GDK_SCREEN_DISPLAY(screen)    (GDK_SCREEN_X11 (screen)->display)
#define GDK_SCREEN_XROOTWIN(screen)   (GDK_SCREEN_X11 (screen)->xroot_window)
#define GDK_WINDOW_SCREEN(win)	      (GDK_DRAWABLE_IMPL_X11 (((GdkWindowObject *)win)->impl)->screen)
#define GDK_WINDOW_DISPLAY(win)       (GDK_SCREEN_X11 (GDK_WINDOW_SCREEN (win))->display)
#define GDK_WINDOW_XROOTWIN(win)      (GDK_SCREEN_X11 (GDK_WINDOW_SCREEN (win))->xroot_window)
#define GDK_GC_DISPLAY(gc)            (GDK_SCREEN_DISPLAY (GDK_GC_X11(gc)->screen))

#include <glib-2.0/glib.h>

#ifdef HAVE_LASH
#include <lash/lash.h>
extern lash_client_t *lash_client;
#endif

gboolean g_quit = FALSE;
gboolean quit = FALSE;

static	GtkWidget* window;
static	GtkWidget* gtk_socket;
static	GtkWidget* vpacker;
static	GtkWidget* hpacker;
static	GtkWidget* bypass_button;
static	GtkWidget* remove_button;
static	GtkWidget* mute_button;
static  GtkWidget* event_box;
static  GtkWidget* preset_listbox;

static void
bypass_handler (GtkToggleButton *but, gboolean ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	
	jvst->bypassed = gtk_toggle_button_get_active (but);
}

static void
mute_handler (GtkToggleButton *but, gboolean ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	jvst->muted = gtk_toggle_button_get_active (but);
}

static void
remove_handler (GtkToggleButton *but, gboolean ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	
	jack_deactivate (jvst->client);
	fst_destroy_editor (jvst->fst);
}

static gboolean
configure_handler (GtkWidget* widget, GdkEventConfigure* ev, GtkSocket *sock)
{
	XEvent event;
	gint x, y;
	GdkWindow *w;

	g_return_if_fail (sock->plug_window != NULL);

	w = sock->plug_window;
	event.xconfigure.type = ConfigureNotify;

	event.xconfigure.event = GDK_WINDOW_XWINDOW (w);
	event.xconfigure.window = GDK_WINDOW_XWINDOW (w);

	/* The ICCCM says that synthetic events should have root relative
	 * coordinates. We still aren't really ICCCM compliant, since
	 * we don't send events when the real toplevel is moved.
	 */
	gdk_error_trap_push ();
	gdk_window_get_origin (w, &x, &y);
	gdk_error_trap_pop ();

	event.xconfigure.x = x;
	event.xconfigure.y = y;
	event.xconfigure.width = GTK_WIDGET(sock)->allocation.width;
	event.xconfigure.height = GTK_WIDGET(sock)->allocation.height;

	event.xconfigure.border_width = 0;
	event.xconfigure.above = None;
	event.xconfigure.override_redirect = False;

	gdk_error_trap_push ();
	XSendEvent (GDK_WINDOW_XDISPLAY (w),
		    GDK_WINDOW_XWINDOW (sock->plug_window),
		    False, StructureNotifyMask, &event);
	//gdk_display_sync (gtk_widget_get_display (GTK_WIDGET (sock)));
	gdk_error_trap_pop ();

	return FALSE;
}

void
forward_key_event (GdkEventKey* ev, GtkSocket* sock, JackVST* jvst)
{
	XKeyEvent event;
	Status status;

	g_return_if_fail (sock->plug_window != NULL);
	
	event.type = (ev->type == GDK_KEY_PRESS ? KeyPress : KeyRelease);
	event.display = GDK_WINDOW_XDISPLAY (sock->plug_window);
	event.window = fst_get_XID (jvst->fst);
	event.time = ev->time;
	event.x = 1;
	event.y = 1;
	event.x_root = 1;
	event.y_root = 1;
	event.state = ev->state;
	event.keycode = ev->hardware_keycode;
	event.same_screen = True;
	
	gdk_error_trap_push ();
	XSendEvent (event.display, event.window, False, 0, (XEvent*) &event);
	gdk_display_sync (gtk_widget_get_display (GTK_WIDGET (sock)));
	gdk_error_trap_pop ();
}

static gboolean
destroy_handler (GtkWidget* widget, GdkEventAny* ev, gpointer ptr)
{
	JackVST* jvst = (JackVST*) ptr;
	fst_destroy_editor (jvst->fst);
	//exit (0);
	gtk_main_quit();
	
	return FALSE;
}

int
focus_handler (GtkWidget* widget, GdkEventFocus* ev, gpointer ptr)
{
	if (ev->in) {
		fst_error ("Socket focus in");
	} else {
		fst_error ("Socket focus out");
	}
		       
	return FALSE;
}

static void
program_change (GtkComboBox *combo, JackVST *jvst)
{
	int program = gtk_combo_box_get_active (combo);
	printf ("active: %d\n", program );
	// cant be done here. plugin only expects one GUI thread.
	//jvst->fst->plugin->dispatcher( jvst->fst->plugin, effSetProgram, 0, program, NULL, 0.0 );
	jvst->fst->want_program = program;
}

#ifdef HAVE_LASH
void
save_data( JackVST *jvst )
{
	int i, bytelen;
	lash_config_t *config;
	void *chunk;

	for( i=0; i<jvst->fst->plugin->numParams; i++ ) {
	    char buf[10];
	    
	    snprintf( buf, 9, "%d", i );

	    config = lash_config_new_with_key( buf );
	    lash_config_set_value_double(config, jvst->fst->plugin->getParameter( jvst->fst->plugin, i ) );
	    lash_send_config(lash_client, config);
	    //lash_config_destroy( config );
	}

	if( jvst->fst->plugin->flags & 32 ) {
	    printf( "getting chunk...\n" );
	    bytelen = jvst->fst->plugin->dispatcher( jvst->fst->plugin, 23, 0, 0, &chunk, 0 );
	    printf( "got tha chunk..\n" );
	    if( bytelen ) {
		if( bytelen < 0 ) {
		    printf( "Chunke len < 0 !!! Not saving chunk.\n" );
		} else {
		    config = lash_config_new_with_key( "bin_chunk" );
		    lash_config_set_value(config, chunk, bytelen );
		    lash_send_config(lash_client, config);
		    //lash_config_destroy( config );
		}


	    }
	}

}

void
restore_data(lash_config_t * config, JackVST *jvst )
{
	const char *key;

	key = lash_config_get_key(config);

	if (strcmp(key, "bin_chunk") == 0) {
	    jvst->fst->plugin->dispatcher( jvst->fst->plugin, 24, 0, lash_config_get_value_size( config ), (void *) lash_config_get_value( config ), 0 );
	    return;
	}

	jvst->fst->plugin->setParameter( jvst->fst->plugin, atoi( key ), lash_config_get_value_double( config ) );

}
#endif

static gboolean
idle_cb(JackVST *jvst)
{
	if (quit) {
		fst_destroy_editor( jvst->fst);
		//gtk_main_quit();
		g_quit = TRUE;
		return FALSE;
	}

#ifdef HAVE_LASH
	if (lash_enabled(lash_client)) {
	    lash_event_t *event;
	    lash_config_t *config;

	    while ((event = lash_get_event(lash_client))) {
		switch (lash_event_get_type(event)) {
		    case LASH_Quit:
			quit = 1;
			lash_event_destroy(event);
			break;
		    case LASH_Restore_Data_Set:
			printf( "lash_restore... \n" );
			lash_send_event(lash_client, event);
			break;
		    case LASH_Save_Data_Set:
			printf( "lash_save... \n" );
			save_data( jvst );
			lash_send_event(lash_client, event);
			break;
		    case LASH_Server_Lost:
			return 1;
		    default:
			printf("%s: receieved unknown LASH event of type %d",
				__FUNCTION__, lash_event_get_type(event));
			lash_event_destroy(event);
			break;
		}
	    }

	    while ((config = lash_get_config(lash_client))) {
		restore_data(config, jvst);
		lash_config_destroy(config);
	    }

	}
#endif
	return TRUE;
}

GtkListStore *create_preset_store( FST *fst )
{
	GtkListStore *retval = gtk_list_store_new( 2, G_TYPE_STRING, G_TYPE_INT );
	int i;
	int vst_version = fst->plugin->dispatcher (fst->plugin, effGetVstVersion, 0, 0, NULL, 0.0f);
	for( i=0; i<fst->plugin->numPrograms; i++ ) {
		char buf[100];
		//char *buf = malloc( 100 );
		GtkTreeIter new_row_iter;
		//GValue int_val = {0};
		//GValue string_val = {0};

		//g_value_init( &int_val, G_TYPE_INT );
		//g_value_init( &string_val, G_TYPE_STRING );

		//fst->plugin->dispatcher( fst->plugin, effSetProgram, 0, i, NULL, 0.0 );

		if( vst_version >= 2 )
			fst->plugin->dispatcher( fst->plugin, 29, i, 0, buf, 0.0 );
		else
			snprintf( buf, 90, "%d", i );

		printf( "program name: %s\n", buf );
		gtk_list_store_insert( retval, &new_row_iter, i );
		//g_value_set_int( &int_val, i );
		//g_value_set_string( &string_val, buf );
		gtk_list_store_set( retval, &new_row_iter, 0, buf, 1, i, -1 );
		//gtk_list_store_set( retval, &new_row_iter, 1, &int_val, -1 );
	}

	if( fst->plugin->numPrograms > 0 )
		fst->plugin->dispatcher( fst->plugin, effSetProgram, 0, 0, NULL, 0.0 );

	return retval;
}
int
manage_vst_plugin (JackVST* jvst)
{
	// create a GtkWindow containing a GtkSocket...
	//
	// notice the order of the functions.
	// you can only add an id to an anchored widget.
	GtkCellRenderer *renderer;
	
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW(window), jvst->handle->name);

	vpacker = gtk_vbox_new (FALSE, 7);
	hpacker = gtk_hbox_new (FALSE, 7);
	bypass_button = gtk_toggle_button_new_with_label ("bypass");
	mute_button = gtk_toggle_button_new_with_label ("mute");
	remove_button = gtk_toggle_button_new_with_label ("remove");
	preset_listbox = gtk_combo_box_new_with_model( GTK_TREE_MODEL(create_preset_store( jvst->fst )) );

	renderer = gtk_cell_renderer_text_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (preset_listbox), renderer, TRUE);
	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (preset_listbox), renderer, "text", 0, NULL);
	gtk_combo_box_set_active( GTK_COMBO_BOX(preset_listbox), 0 );
	g_signal_connect( G_OBJECT(preset_listbox), "changed", G_CALLBACK( program_change ), jvst ); 






	gtk_signal_connect (GTK_OBJECT(bypass_button), "toggled",
			    GTK_SIGNAL_FUNC(bypass_handler),
			    jvst);

	gtk_signal_connect (GTK_OBJECT(mute_button), "toggled",
			    GTK_SIGNAL_FUNC(mute_handler),
			    jvst);

	gtk_signal_connect (GTK_OBJECT(remove_button), "toggled",
			    GTK_SIGNAL_FUNC(remove_handler),
			    jvst);

	gtk_container_set_border_width (GTK_CONTAINER(hpacker), 3);

	gtk_signal_connect (GTK_OBJECT(window), "delete_event",
			    GTK_SIGNAL_FUNC(destroy_handler),
			    jvst);

	gtk_socket = gtk_socket_new ();
	GTK_WIDGET_SET_FLAGS(gtk_socket, GTK_CAN_FOCUS);
	
	gtk_box_pack_end   (GTK_BOX(hpacker), preset_listbox, FALSE, FALSE, 0);
	gtk_box_pack_end   (GTK_BOX(hpacker), bypass_button, FALSE, FALSE, 0);
	gtk_box_pack_end   (GTK_BOX(hpacker), mute_button, FALSE, FALSE, 0);
	// gtk_box_pack_end   (GTK_BOX(hpacker), remove_button, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vpacker), hpacker, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX(vpacker), gtk_socket, TRUE, TRUE, 0);

	gtk_container_add (GTK_CONTAINER (window), vpacker);

	// normally every socket should register it self like this.
	gtk_signal_connect (GTK_OBJECT(window), "configure_event",
//	gtk_signal_connect (GTK_OBJECT(gtk_socket), "configure_event",
			    GTK_SIGNAL_FUNC(configure_handler),
			    gtk_socket);

	//gtk_socket_steal (GTK_SOCKET (gtk_socket), fst_get_XID (jvst->fst));
	
	//fst_show_editor (jvst->fst);

	// but you can show() a GtkSocket only with an id set.
	gtk_socket_add_id (GTK_SOCKET (gtk_socket), fst_get_XID (jvst->fst));
	
 	gtk_widget_show_all (window);
	//gtk_window_set_focus (GTK_WINDOW(window), gtk_socket);

	SetWindowPos (jvst->fst->window, 0, 0, 0, jvst->fst->width, jvst->fst->height, 0);
	ShowWindow (jvst->fst->window, SW_SHOWNA);


	g_timeout_add(500, (GSourceFunc) idle_cb, jvst);

	printf( "calling gtk_main now\n" );
	gtk_main ();

	return 0;
}

/*
int
manage_vst_plugin (JackVST* jvst)
{
	while( !g_quit ) {
	    idle_cb( jvst );
	    //usleep( 500 );
	}

	return 0;
}
*/

typedef int (*error_handler_t)( Display *, XErrorEvent *);
void
gui_init (int *argc, char **argv[])
{
	error_handler_t errhandle = XSetErrorHandler( NULL );
	gtk_init (argc, argv);
	XSetErrorHandler( errhandle );
}
