/*
    Copyright (C) 2004 Paul Davis <paul@linuxaudiosystems.com> 
                       Torben Hohn <torbenh@informatik.uni-bremen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <fst.h>
#include <glib.h>
#include <semaphore.h>
#include <vestige/aeffectx.h>

#include "jackvst.h"
#ifdef HAVE_LASH
#include <lash/lash.h>
#endif
#include <windows.h>

/* audiomaster.c */

extern long jack_host_callback (struct AEffect*, long, long, long, void*, float);

/* gtk.c */

extern void gui_init (int* argc, char** argv[]);
extern int  manage_vst_plugin (JackVST*);

/* vsti.c */

extern snd_seq_t * create_sequencer (const char* client_name, int isinput);
extern void *midireceiver(void *arg);

#ifdef HAVE_LASH
lash_client_t * lash_client;
#endif


// Until i get Jack to call me when he wants to create a thread,
// i need to go through this.

volatile int with_winaudio = 1;

static HANDLE audioThreadHandle = 0;
static jack_nframes_t winaudio_nframes;
static void *winaudio_data;
sem_t winaudio_activate, winaudio_done;

int winaudio_process_callback( jack_nframes_t nframes, void *data ) {
    winaudio_nframes = nframes;
    winaudio_data = data;

    sem_post( &winaudio_activate );
    sem_wait( &winaudio_done );

    return 0;
}

DWORD WINAPI
WinAudioThreadMain(LPVOID parameter) {

    struct sched_param param;
    param.sched_priority = 10;
    pthread_setschedparam(pthread_self(), SCHED_FIFO, &param);

    sem_post( &winaudio_done );

    while( with_winaudio ) {
	sem_wait( &winaudio_activate );
	process_callback( winaudio_nframes, winaudio_data );
	sem_post( &winaudio_done );
    }
    return 0;
}

int process_callback( jack_nframes_t nframes, void* data) 
{
	int i, o;
	JackVST* jvst = (JackVST*) data;
	struct AEffect* plugin = jvst->fst->plugin;

	if( !jvst->resume_called ) {
	    jvst->resume_called = TRUE;
	    plugin->dispatcher (plugin, effMainsChanged, 0, 1, NULL, 0.0f);
	}
	for (i = 0; i < plugin->numInputs; ++i) {
		jvst->ins[i]  = (float *) jack_port_get_buffer (jvst->inports[i], nframes);
	}

	for (i = 0; i < plugin->numOutputs; ++i) {
		jvst->outs[i]  = (float *) jack_port_get_buffer (jvst->outports[i], nframes);
	}

	if (jvst->bypassed) {

		if (plugin->numInputs) {
			for (o = 0, i = 0; o < plugin->numOutputs; ++o) {
				memcpy (jvst->outs[o], jvst->ins[i], sizeof (float) * nframes);
				
				if (i < plugin->numOutputs - 1) {
					++i;
				}
			}
		} 
		
	} else if (jvst->muted) {

		for (o = 0, i = 0; o < plugin->numOutputs; ++o) {
			if (jvst->outs[o]) {
				memset (jvst->outs[o], 0, sizeof (float) * nframes);
			}
		}
		
	} else {

		if (jvst->event_queue) {

			jack_ringbuffer_data_t vec[2];
			size_t n;
			size_t i;

			jack_ringbuffer_get_read_vector (jvst->event_queue, vec);

			if ((jvst->events->numEvents = (vec[0].len / sizeof (struct VstMidiEvent))) > 0) {
				jvst->events->reserved  = 0;
				
				for (i = 0, n = 0; n < jvst->events->numEvents; ++n, ++i) {
					jvst->events->events[i] = (struct VstEvent*) (vec[0].buf + (n * sizeof (struct VstMidiEvent)));
				}

				plugin->dispatcher (plugin, effProcessEvents, 0, 0, jvst->events, 0.0f);
			}

			if (vec[1].len && (jvst->events->numEvents = (vec[1].len / sizeof (struct VstMidiEvent)) > 0)) {
				jvst->events->reserved  = 0;

				/* don't initialize i again, continue stuffing events after
				   the last batch.
				 */
				
				for (n = 0; n < jvst->events->numEvents; ++n, ++i) {
					jvst->events->events[i] = (struct VstEvent*) (vec[0].buf + (n * sizeof (struct VstMidiEvent)));
				}

				plugin->dispatcher (plugin, effProcessEvents, 0, 0, jvst->events, 0.0f);
			}
			
			jack_ringbuffer_read_advance (jvst->event_queue, (vec[0].len + vec[1].len));
		}


		if (plugin->flags & effFlagsCanReplacing) {
			
			for (i = 0; i < plugin->numOutputs; ++i) {
				memset (jvst->outs[i], 0, sizeof (float) * nframes);
			}
			plugin->processReplacing (plugin, jvst->ins, jvst->outs, nframes);
			
		} else {
			
			for (i = 0; i < plugin->numOutputs; ++i) {
				memset (jvst->outs[i], 0, sizeof (float) * nframes);
			}
			plugin->process (plugin, jvst->ins, jvst->outs, nframes);
		}
	}

	return 0;      
}

enum ParseMode {
    MODE_NORMAL,
    MODE_QUOTE,
    MODE_DOUBLEQUOTE,
    MODE_ESCAPED,
    MODE_WHITESPACE,
    MODE_EOL

};
void create_argc_argv_from_cmdline( char *cmdline, char *argv0, int *argc, char ***argv ) {
    // first count argc
    char *pos = cmdline;
    enum ParseMode parseMode = MODE_WHITESPACE;
    enum ParseMode parseMode_before_ESC = MODE_NORMAL;
    int i;

    int myargc = 1; 
    char **myargv;

    while( parseMode != MODE_EOL ) {
	switch( parseMode ) {
	    case MODE_NORMAL:
		switch( *pos ) {
		    case '\"':
			parseMode = MODE_DOUBLEQUOTE;
			break;
		    case '\'':
			parseMode = MODE_QUOTE;
			break;
		    case '\\':
			parseMode_before_ESC = parseMode;
			parseMode = MODE_ESCAPED;
			break;
		    case ' ':	// First Space after an arg;
			parseMode = MODE_WHITESPACE;
			myargc++;
			break;
		    case 0:	// EOL after arg.
			parseMode = MODE_EOL;
			break;
		    default:
			// Normal char;
			break;
		}
		break;
	    case MODE_QUOTE:
		switch( *pos ) {
		    case '\'':
			parseMode = MODE_NORMAL;
			break;
		    case '\\':
			parseMode_before_ESC = parseMode;
			parseMode = MODE_ESCAPED;
			break;
		    case 0:
			fst_error( "parse Error on cmdline" );
			parseMode = MODE_EOL;
			break;
		    default:
			// Normal char;
			break;
		}
		break;
	    case MODE_DOUBLEQUOTE:
		switch( *pos ) {
		    case '\"':
			parseMode = MODE_NORMAL;
			break;
		    case '\\':
			parseMode_before_ESC = parseMode;
			parseMode = MODE_ESCAPED;
			break;
		    case 0:
			fst_error( "parse Error on cmdline" );
			parseMode = MODE_EOL;
			break;
		    default:
			// Normal char;
			break;
		}
		break;
	    case MODE_ESCAPED:
		switch( *pos ) {
		    case '\"':
			// emit escaped char;
			parseMode = parseMode_before_ESC;
			break;
		    case '\'':
			// emit escaped char;
			parseMode = parseMode_before_ESC;
			break;
		    case '\\':
			// emit escaped char;
			parseMode = parseMode_before_ESC;
			break;
		    case 0:
			fst_error( "EOL after escape: ignored" );
			parseMode = MODE_EOL;
			break;
		    default:
			fst_error( "Unknown Escapecharacter: ignored" );
			parseMode = parseMode_before_ESC;
			// Normal char;
			break;
		}
		break;
	    case MODE_WHITESPACE:
		switch( *pos ) {
		    case '\"':
			parseMode = MODE_DOUBLEQUOTE;
			myargc++;
			break;
		    case '\'':
			parseMode = MODE_QUOTE;
			myargc++;
			break;
		    case '\\':
			parseMode_before_ESC = MODE_NORMAL;
			parseMode = MODE_ESCAPED;
			myargc++;
			break;
		    case ' ':
			parseMode = MODE_WHITESPACE;
			break;
		    case 0:
			parseMode = MODE_EOL;
			break;
		    default:
			// 
			parseMode = MODE_NORMAL;
			// Normal char;
			myargc++;
			break;
		}
		break;
	}
	pos++;
    }

    myargv = malloc( myargc * sizeof( char * ) );
    if( !myargv ) {
	fst_error( "cant alloc memory" );
	exit( 10 );
    }

    // alloc strlen(cmdline) + 1 for each argv.
    // this avoids another parsing pass.
    for( i=0; i<myargc; i++ ) {
	myargv[i] = malloc( strlen(cmdline) + 1 );
	if( !myargv[i] ) {
	    fst_error( "cant alloc memory" );
	    exit( 10 );
	}
	myargv[i][0] = 0;
    }

    // Now rerun theparser and actually emit chars.
    pos = cmdline;
    parseMode = MODE_WHITESPACE;
    parseMode_before_ESC = MODE_NORMAL;
    int current_arg = 0;
    char *emit_pos = myargv[0];
    
    while( parseMode != MODE_EOL ) {
	switch( parseMode ) {
	    case MODE_NORMAL:
		switch( *pos ) {
		    case '\"':
			parseMode = MODE_DOUBLEQUOTE;
			break;
		    case '\'':
			parseMode = MODE_QUOTE;
			break;
		    case '\\':
			parseMode_before_ESC = parseMode;
			parseMode = MODE_ESCAPED;
			break;
		    case ' ':	// First Space after an arg;
			parseMode = MODE_WHITESPACE;
			*emit_pos = 0;

			break;
		    case 0:	// EOL after arg.
			parseMode = MODE_EOL;
		        *emit_pos = 0;	
			break;
		    default:

			*(emit_pos++) = *pos;
			break;
		}
		break;
	    case MODE_QUOTE:
		switch( *pos ) {
		    case '\'':
			parseMode = MODE_NORMAL;
			break;
		    case '\\':
			parseMode_before_ESC = parseMode;
			parseMode = MODE_ESCAPED;
			break;
		    case 0:
			fst_error( "parse Error on cmdline" );
			parseMode = MODE_EOL;
		        *emit_pos = 0;	
			break;
		    default:
			// Normal char;
			*(emit_pos++) = *pos;
			break;
		}
		break;
	    case MODE_DOUBLEQUOTE:
		switch( *pos ) {
		    case '"':
			parseMode = MODE_NORMAL;
			break;
		    case '\\':
			parseMode_before_ESC = parseMode;
			parseMode = MODE_ESCAPED;
			break;
		    case 0:
			fst_error( "parse Error on cmdline" );
			parseMode = MODE_EOL;
		        *emit_pos = 0;	
			break;
		    default:
			// Normal char;
			*(emit_pos++) = *pos;
			break;
		}
		break;
	    case MODE_ESCAPED:
		switch( *pos ) {
		    case '\"':
			// emit escaped char;
			parseMode = parseMode_before_ESC;
			*(emit_pos++) = *pos;
			break;
		    case '\'':
			// emit escaped char;
			parseMode = parseMode_before_ESC;
			*(emit_pos++) = *pos;
			break;
		    case '\\':
			// emit escaped char;
			parseMode = parseMode_before_ESC;
			*(emit_pos++) = *pos;
			break;
		    case 0:
			fst_error( "EOL after escape: ignored" );
			parseMode = MODE_EOL;
		        *emit_pos = 0;	
			break;
		    default:
			fst_error( "Unknown Escapecharacter: ignored" );
			parseMode = parseMode_before_ESC;
			break;
		}
		break;
	    case MODE_WHITESPACE:
		switch( *pos ) {
		    case '\"':
			parseMode = MODE_DOUBLEQUOTE;
			// ok... arg begins
			current_arg++;
			emit_pos = myargv[current_arg];
			break;
		    case '\'':
			parseMode = MODE_QUOTE;
			// ok... arg begins
			current_arg++;
			emit_pos = myargv[current_arg];
			break;
		    case '\\':
			parseMode_before_ESC = MODE_NORMAL;
			parseMode = MODE_ESCAPED;
			// ok... arg begins
			current_arg++;
			emit_pos = myargv[current_arg];
			break;
		    case ' ':
			parseMode = MODE_WHITESPACE;
			break;
		    case 0:
			parseMode = MODE_EOL;
			break;
		    default:
			parseMode = MODE_NORMAL;
			// ok... arg begins
			current_arg++;
			emit_pos = myargv[current_arg];
			// Normal char;
			*(emit_pos++) = *pos;
			break;
		}
		break;
	}
	pos++;
    }

    strncpy( myargv[0], argv0, strlen(cmdline) );
    
    *argc = myargc;
    *argv = myargv;
}

int WINAPI
WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR cmdline, int cmdshow)
//int 
//main( int argc, char **argv ) 
{
	JackVST* jvst;
	struct AEffect* plugin;
	int i;
	char* client_name;
	char* period;
	int with_editor = 1;
	int resume_not_rt = 1;
	char *plug;
	int vst_version;

	int argc;
	char **argv;

	create_argc_argv_from_cmdline( cmdline, "fst", &argc, &argv );

#ifdef HAVE_LASH

	lash_event_t *event;
	lash_args_t *lash_args;

	if ((period = strrchr (argv[0], '.')) != NULL) {
		*period = '\0';
	}
	if ((period = strrchr (argv[0], '.')) != NULL) {
		*period = '\0';
	}
	lash_args = lash_extract_args(&argc, &argv);
#endif

	if (argc < 2) {
		fprintf (stderr, "usage: %s <plugin>\n", argv[0]);
		return 1;
	}

	printf( "yo... lets see...\n" );
	//setpgrp();

	gui_init (&argc, &argv);
#if 1
	for (i = 1; i < argc; ++i) {
		if (argv[i][0] == '-') {
			if (argv[i][1] == 'n') {
				with_editor = 0;
			}
			if (argv[i][1] == 'r') {
				resume_not_rt = 0;
			}
		} else {
			plug = argv[i];
			break;
		}
	}
#else
	printf( "cmdline: >%s<\n", cmdline );
	plug = cmdline;
#endif

	if (fst_init (hInst)) {
		return 1;
	}

	jvst = (JackVST*) calloc (1, sizeof (JackVST));

	client_name = g_path_get_basename(strdup (plug));
	if ((period = strrchr (client_name, '.')) != NULL) {
		*period = '\0';
	}


	if ((jvst->handle = fst_load (plug)) == NULL) {
		fst_error ("can't load plugin %s", plug);
		return 1;
	}
	if ((jvst->client = jack_client_open (client_name, JackNullOption, NULL )) == 0) {
		fst_error ("can't connect to JACK");
		return 1;
	}

	printf( "instantiate... \n" );

	if ((jvst->fst = fst_instantiate (jvst->handle, jack_host_callback, jvst)) == NULL) {
		fst_error ("can't instantiate plugin %s", plug);
		return 1;
	}

	plugin = jvst->fst->plugin;
	
	/* set rate and blocksize */

	printf( "sample_rate... \n" );

	plugin->dispatcher (plugin, effSetSampleRate, 0, 0, NULL, 
			    (float) jack_get_sample_rate (jvst->client));
	plugin->dispatcher (plugin, effSetBlockSize, 0, 
			    jack_get_buffer_size (jvst->client), NULL, 0.0f);

	if( resume_not_rt ) {
	    jvst->resume_called = TRUE;
	    plugin->dispatcher (plugin, effMainsChanged, 0, 1, NULL, 0.0f);
	}
	
	// ok.... plugin is running... lets bind to lash...
	
#ifdef HAVE_LASH
	int flags = LASH_Config_Data_Set;

	lash_client =
		lash_init(lash_args, client_name, flags, LASH_PROTOCOL(2, 0));

	if (!lash_client) {
	    fprintf(stderr, "%s: could not initialise lash\n", __FUNCTION__);
	    fprintf(stderr, "%s: running fst without lash session-support\n", __FUNCTION__);
	    fprintf(stderr, "%s: to enable lash session-support launch the lash server prior fst\n", __FUNCTION__);
	    //exit(1);
	}

	if (lash_enabled(lash_client)) {
		event = lash_event_new_with_type(LASH_Client_Name);
		lash_event_set_string(event, client_name);
		lash_send_event(lash_client, event);
	}
#endif


	/* set program to zero */
	/* i comment this out because it breaks dfx Geometer
	 * looks like we cant set programs for it
	 *
	plugin->dispatcher (plugin, effSetProgram, 0, 0, NULL, 0.0f); */

	vst_version = plugin->dispatcher (plugin, effGetVstVersion, 0, 0, NULL, 0.0f);

	if (vst_version >= 2) {
		
                /* should we send it VST events (i.e. MIDI) */
		
		if ((plugin->flags & effFlagsIsSynth) ||
		    (plugin->dispatcher (plugin, effCanDo, 0, 0,(void*) "receiveVstEvents", 0.0f) > 0)) {

			if ((jvst->seq = create_sequencer (jvst->handle->name, TRUE)) == NULL) {
				fst_error ("no sequencer for synth");
			} else {
				jvst->events = (struct VstEvents*) malloc (sizeof (struct VstEvents) - (2*sizeof (struct VstEvent*)) +
									   (1024 * sizeof (struct VstMidiEvent*)));

				jvst->event_queue = jack_ringbuffer_create (1024 * sizeof (struct VstMidiEvent));
				pthread_create (&jvst->midi_thread, NULL, &midireceiver, jvst);
			}
#ifdef HAVE_LASH
			if( lash_enabled( lash_client ) )
			    lash_alsa_client_id(lash_client, (unsigned char)snd_seq_client_id(jvst->seq));
#endif
		}
	}

	printf( "PortLayout: in: %d out: %d\n", plugin->numInputs, plugin->numOutputs );
	jvst->inports = (jack_port_t **) malloc (sizeof(jack_port_t*) * plugin->numInputs);
	jvst->ins = (float **) malloc (sizeof (float *) * plugin->numInputs);
	
	for (i = 0; i < plugin->numInputs; ++i) {
		char buf[64];
		snprintf (buf, sizeof(buf), "in%d", i+1);
		jvst->inports[i] = jack_port_register (jvst->client, buf, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	}
	
	jvst->outports = (jack_port_t **) malloc (sizeof(jack_port_t*) * plugin->numOutputs);
	jvst->outs = (float **) malloc (sizeof (float *) * plugin->numOutputs);
	
	for (i = 0; i < plugin->numOutputs; ++i) {
		char buf[64];
		snprintf (buf, sizeof(buf), "out%d", i+1);
		jvst->outports[i] = jack_port_register (jvst->client, buf, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	}

	if( with_winaudio ) {
	    DWORD threadId = 0;

	    sem_init( &winaudio_activate, 0, 0 );
	    sem_init( &winaudio_done, 0, 0 );
	    
	    audioThreadHandle = CreateThread(0, 0, WinAudioThreadMain, 0, 0, &threadId);
	    if (!audioThreadHandle) {
		fst_error( "Failed to create audio thread!" );
		return 1;
	    }
	    // ok... this waits for the windows thread to get up in the first place.
	    sem_wait( &winaudio_done );
	    printf( "created audio thread\n" );
	    jack_set_process_callback (jvst->client, (JackProcessCallback) winaudio_process_callback, jvst); 
	} else {
	
	    jack_set_process_callback (jvst->client, (JackProcessCallback) process_callback, jvst); 
	}
	printf( "Calling Jack activate\n" );
	jack_activate (jvst->client);

	if (with_editor) {
		if (fst_run_editor (jvst->fst)) {
			fst_error ("cannot create editor");
			return 1;
		}
	}

#ifdef HAVE_LASH
	if( lash_enabled( lash_client ) ) {
	    event = lash_event_new_with_type(LASH_Jack_Client_Name);
	    lash_event_set_string(event, client_name);
	    lash_send_event(lash_client, event);
	}
#endif

	printf( "Entering main loop\n" );
	if (with_editor) {
		printf( "hallo ?\n" );
		manage_vst_plugin (jvst);
		//gui_event_loop(jvst->fst);
	//    while( 1 )
	//	sleep (10);
	} else {
	    while( 1 )
		sleep (10);
	}
	
	jack_deactivate( jvst->client );
	if( with_winaudio ) {
	    printf( "Waiting for WinAudioThread to exit\n" );
	    with_winaudio = 0;
	    CloseHandle( audioThreadHandle );
	}
	return 0;
}
